from db import get_database
from datetime import datetime, timedelta
from dateutil.relativedelta import relativedelta

def list_players_contract_expiring_next_12_months():
    db = get_database()
    now = datetime.now()
    one_year_from_now = now + relativedelta(years=1)
    pipeline = [
        {"$match": {"contracts.start_date": {"$lte": one_year_from_now}}},
        {"$unwind": "$contracts"},
        {"$set": {
            "contract_end_date": {
                "$dateAdd": {
                    "startDate": "$contracts.start_date",
                    "unit": "year",
                    "amount": "$contracts.duration_years"
                }
            }
        }},
        {"$match": {
            "contract_end_date": {
                "$lt": one_year_from_now,
                "$gte": now
            }
        }},
        {"$project": {
            "name": 1
        }}
    ]
    return [doc['name'] for doc in db.players.aggregate(pipeline)]


def list_female_players_three_years_service():
    db = get_database()
    now = datetime.now()
    pipeline = [
        {"$match": {"gender": "F"}},
        {"$unwind": "$contracts"},
        {"$addFields": {
            "years_of_service": {
                "$divide": [
                    {"$dateDiff": {
                        "startDate": "$contracts.start_date",
                        "endDate": "$$NOW",
                        "unit": "year"
                    }},
                    1
                ]
            }
        }},
        {"$match": {"years_of_service": {"$gte": 3}}},
        {"$group": {
            "_id": "$_id",
            "name": {"$first": "$name"},
            "max_years_of_service": {"$max": "$years_of_service"}
        }},
        {"$match": {"max_years_of_service": {"$gte": 3}}},
        {"$project": {"name": 1}}
    ]
    return [doc['name'] for doc in db.players.aggregate(pipeline)]


def list_top_5_income_generating_players():
    db = get_database()
    pipeline = [
        {"$unwind": "$contracts"},
        {"$addFields": {
            "income_generation": {
                "$multiply": ["$contracts.salary_per_week", "$contracts.duration_years", 52, "$contracts.commission_percentage", 0.01]
            }
        }},
        {"$sort": {"income_generation": -1}},
        {"$limit": 5},
        {"$project": {"name": 1}}
    ]
    return [doc['name'] for doc in db.players.aggregate(pipeline)]

def list_top_5_longest_associated_players():
    db = get_database()
    pipeline = [
        {"$sort": {"date_signed_up": 1}},
        {"$limit": 5},
        {"$project": {"name": 1}}
    ]
    return [doc['name'] for doc in db.players.aggregate(pipeline)]

def list_players_at_risk_no_contract_renewal():
    db = get_database()
    now = datetime.now()
    six_months_from_now = now + relativedelta(months=6)
    pipeline = [
        {"$match": {"contracts.start_date": {"$lte": six_months_from_now}}},
        {"$unwind": "$contracts"},
        {"$set": {
            "contract_end_date": {
                "$dateAdd": {
                    "startDate": "$contracts.start_date",
                    "unit": "year",
                    "amount": "$contracts.duration_years"
                }
            }
        }},
        {"$match": {
            "contract_end_date": {
                "$lt": six_months_from_now,
                "$gte": now
            }
        }},
        {"$project": {
            "name": 1
        }}
    ]
    return [doc['name'] for doc in db.players.aggregate(pipeline)]



if __name__ == "__main__":
    print("Players with contracts expiring in the next 12 months:")
    for name in list_players_contract_expiring_next_12_months():
        print(f"Name: {name}")

    print("\nFemale players with at least three years of service with the same team:")
    for name in list_female_players_three_years_service():
        print(f"Name: {name}")

    print("\nTop 5 income-generating players:")
    for name in list_top_5_income_generating_players():
        print(f"Name: {name}")

    print("\nTop 5 longest associated players:")
    for name in list_top_5_longest_associated_players():
        print(f"Name: {name}")

    print("\nPlayers at risk of no contract renewal:")
    for name in list_players_at_risk_no_contract_renewal():
        print(f"Name: {name}")
